@@ -18,23 +18,20 @@
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
-# force use of pygobject3 in python-slip
-from gi.repository import GObject
-import sys
-sys.modules['gobject'] = GObject
 import os
 
 import dbus
 import dbus.service
-import slip.dbus
-import slip.dbus.service
 
 from firewall import config
 from firewall.core.base import DEFAULT_ZONE_TARGET
 from firewall.core.watcher import Watcher
 from firewall.core.logger import log
+from firewall.server.dbus import DbusServiceObject
 from firewall.server.decorators import handle_exceptions, \
-    dbus_handle_exceptions, dbus_service_method
+    dbus_handle_exceptions, dbus_service_method, \
+    dbus_service_method_deprecated, dbus_service_signal_deprecated, \
+    dbus_polkit_require_auth
 from firewall.server.config_icmptype import FirewallDConfigIcmpType
 from firewall.server.config_service import FirewallDConfigService
 from firewall.server.config_zone import FirewallDConfigZone
@@ -49,7 +46,8 @@ from firewall.core.io.direct import Direct
 from firewall.dbus_utils import dbus_to_python, \
     command_of_sender, context_of_sender, uid_of_sender, user_of_uid, \
     dbus_introspection_prepare_properties, \
-    dbus_introspection_add_properties
+    dbus_introspection_add_properties, \
+    dbus_introspection_add_deprecated
 from firewall import errors
 from firewall.errors import FirewallError
 
@@ -59,7 +57,7 @@ from firewall.errors import FirewallError
 #
 ############################################################################
 
-class FirewallDConfig(slip.dbus.service.Object):
+class FirewallDConfig(DbusServiceObject):
     """FirewallD main class"""
 
     persistent = True
@@ -97,21 +95,24 @@ class FirewallDConfig(slip.dbus.service.Object):
         self.watcher.add_watch_file(config.FIREWALLD_DIRECT)
         self.watcher.add_watch_file(config.FIREWALLD_CONF)
 
-        dbus_introspection_prepare_properties(self,
-                                              config.dbus.DBUS_INTERFACE_CONFIG,
-                                              { "CleanupOnExit": "readwrite",
-                                                "CleanupModulesOnExit": "readwrite",
-                                                "IPv6_rpfilter": "readwrite",
-                                                "Lockdown": "readwrite",
-                                                "MinimalMark": "readwrite",
-                                                "IndividualCalls": "readwrite",
-                                                "LogDenied": "readwrite",
-                                                "AutomaticHelpers": "readwrite",
-                                                "FirewallBackend": "readwrite",
-                                                "FlushAllOnReload": "readwrite",
-                                                "RFC3964_IPv4": "readwrite",
-                                                "AllowZoneDrifting": "readwrite",
-                                              })
+        dbus_introspection_prepare_properties(self, config.dbus.DBUS_INTERFACE_CONFIG,
+            {
+                "CleanupOnExit": "readwrite",
+                "CleanupModulesOnExit": "readwrite",
+                "IPv6_rpfilter": "readwrite",
+                "IPv6_rpfilter2": "readwrite",
+                "Lockdown": "readwrite",
+                "MinimalMark": "readwrite",
+                "IndividualCalls": "readwrite",
+                "LogDenied": "readwrite",
+                "AutomaticHelpers": "readwrite",
+                "FirewallBackend": "readwrite",
+                "FlushAllOnReload": "readwrite",
+                "RFC3964_IPv4": "readwrite",
+                "AllowZoneDrifting": "readwrite",
+                "NftablesTableOwner": "readwrite",
+            }
+        )
 
     @handle_exceptions
     def _init_vars(self):
@@ -529,6 +530,13 @@ class FirewallDConfig(slip.dbus.service.Object):
 
     @dbus_handle_exceptions
     def accessCheck(self, sender):
+        if self.config._fw._state == "FAILED":
+            raise FirewallError(errors.RUNNING_BUT_FAILED,
+                    "Changing permanent configuration is not allowed while "
+                    "firewalld is in FAILED state. The permanent "
+                    "configuration must be fixed and then firewalld "
+                    "restarted. Try `firewall-offline-cmd --check-config`.")
+
         if self.config.lockdown_enabled():
             if sender is None:
                 log.error("Lockdown not possible, sender not set.")
@@ -558,7 +566,7 @@ class FirewallDConfig(slip.dbus.service.Object):
                          "CleanupModulesOnExit", "Lockdown", "IPv6_rpfilter",
                          "IndividualCalls", "LogDenied", "AutomaticHelpers",
                          "FirewallBackend", "FlushAllOnReload", "RFC3964_IPv4",
-                         "AllowZoneDrifting" ]:
+                         "AllowZoneDrifting", "IPv6_rpfilter2", "NftablesTableOwner" ]:
             raise dbus.exceptions.DBusException(
                 "org.freedesktop.DBus.Error.InvalidArgs: "
                 "Property '%s' does not exist" % prop)
@@ -588,8 +596,13 @@ class FirewallDConfig(slip.dbus.service.Object):
                 value = "yes" if config.FALLBACK_LOCKDOWN else "no"
             return dbus.String(value)
         elif prop == "IPv6_rpfilter":
+            if value is None or value != "no":
+                return dbus.String("yes")
+            else:
+                return dbus.String("no")
+        elif prop == "IPv6_rpfilter2":
             if value is None:
-                value = "yes" if config.FALLBACK_IPV6_RPFILTER else "no"
+                value = config.FALLBACK_IPV6_RPFILTER
             return dbus.String(value)
         elif prop == "IndividualCalls":
             if value is None:
@@ -619,6 +632,10 @@ class FirewallDConfig(slip.dbus.service.Object):
             if value is None:
                 value = "yes" if config.FALLBACK_ALLOW_ZONE_DRIFTING else "no"
             return dbus.String(value)
+        elif prop == "NftablesTableOwner":
+            if value is None:
+                value = "yes" if config.FALLBACK_NFTABLES_TABLE_OWNER else "no"
+            return dbus.String(value)
 
     @dbus_handle_exceptions
     def _get_dbus_property(self, prop):
@@ -634,6 +651,8 @@ class FirewallDConfig(slip.dbus.service.Object):
             return dbus.String(self._get_property(prop))
         elif prop == "IPv6_rpfilter":
             return dbus.String(self._get_property(prop))
+        elif prop == "IPv6_rpfilter2":
+            return dbus.String(self._get_property(prop))
         elif prop == "IndividualCalls":
             return dbus.String(self._get_property(prop))
         elif prop == "LogDenied":
@@ -648,6 +667,8 @@ class FirewallDConfig(slip.dbus.service.Object):
             return dbus.String(self._get_property(prop))
         elif prop == "AllowZoneDrifting":
             return dbus.String(self._get_property(prop))
+        elif prop == "NftablesTableOwner":
+            return dbus.String(self._get_property(prop))
         else:
             raise dbus.exceptions.DBusException(
                 "org.freedesktop.DBus.Error.InvalidArgs: "
@@ -674,8 +695,6 @@ class FirewallDConfig(slip.dbus.service.Object):
                 "org.freedesktop.DBus.Error.UnknownInterface: "
                 "Interface '%s' does not exist" % interface_name)
 
-        return self._get_dbus_property(property_name)
-
     @dbus_service_method(dbus.PROPERTIES_IFACE, in_signature='s',
                          out_signature='a{sv}')
     @dbus_handle_exceptions
@@ -689,7 +708,7 @@ class FirewallDConfig(slip.dbus.service.Object):
                        "CleanupModulesOnExit", "Lockdown", "IPv6_rpfilter",
                        "IndividualCalls", "LogDenied", "AutomaticHelpers",
                        "FirewallBackend", "FlushAllOnReload", "RFC3964_IPv4",
-                       "AllowZoneDrifting" ]:
+                       "AllowZoneDrifting", "IPv6_rpfilter2", "NftablesTableOwner" ]:
                 ret[x] = self._get_property(x)
         elif interface_name in [ config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                                  config.dbus.DBUS_INTERFACE_CONFIG_POLICIES ]:
@@ -701,7 +720,7 @@ class FirewallDConfig(slip.dbus.service.Object):
 
         return dbus.Dictionary(ret, signature="sv")
 
-    @slip.dbus.polkit.require_auth(config.dbus.PK_ACTION_CONFIG)
+    @dbus_polkit_require_auth(config.dbus.PK_ACTION_CONFIG)
     @dbus_service_method(dbus.PROPERTIES_IFACE, in_signature='ssv')
     @dbus_handle_exceptions
     def Set(self, interface_name, property_name, new_value, sender=None):
@@ -713,49 +732,48 @@ class FirewallDConfig(slip.dbus.service.Object):
         self.accessCheck(sender)
 
         if interface_name == config.dbus.DBUS_INTERFACE_CONFIG:
-            if property_name in [ "CleanupOnExit", "Lockdown", "CleanupModulesOnExit",
+            if property_name in [ "CleanupOnExit",
+                                  "CleanupModulesOnExit", "Lockdown",
                                   "IPv6_rpfilter", "IndividualCalls",
                                   "LogDenied",
                                   "FirewallBackend", "FlushAllOnReload",
-                                  "RFC3964_IPv4", "AllowZoneDrifting" ]:
-                if property_name in [ "CleanupOnExit", "Lockdown", "CleanupModulesOnExit",
-                                      "IPv6_rpfilter", "IndividualCalls" ]:
+                                  "RFC3964_IPv4", "IPv6_rpfilter2",
+                                  "NftablesTableOwner" ]:
+                if property_name in [ "CleanupOnExit", "CleanupModulesOnExit",
+                                      "Lockdown", "IPv6_rpfilter",
+                                      "IndividualCalls", "FlushAllOnReload",
+                                      "RFC3964_IPv4", "NftablesTableOwner" ]:
                     if new_value.lower() not in [ "yes", "no",
                                                   "true", "false" ]:
                         raise FirewallError(errors.INVALID_VALUE,
                                             "'%s' for %s" % \
                                             (new_value, property_name))
-                if property_name == "LogDenied":
+                elif property_name == "LogDenied":
                     if new_value not in config.LOG_DENIED_VALUES:
                         raise FirewallError(errors.INVALID_VALUE,
                                             "'%s' for %s" % \
                                             (new_value, property_name))
-                if property_name == "FirewallBackend":
+                elif property_name == "FirewallBackend":
                     if new_value not in config.FIREWALL_BACKEND_VALUES:
                         raise FirewallError(errors.INVALID_VALUE,
                                             "'%s' for %s" % \
                                             (new_value, property_name))
-                if property_name == "FlushAllOnReload":
-                    if new_value.lower() not in ["yes", "true", "no", "false"]:
-                        raise FirewallError(errors.INVALID_VALUE,
-                                            "'%s' for %s" % \
-                                            (new_value, property_name))
-                if property_name == "RFC3964_IPv4":
-                    if new_value.lower() not in ["yes", "true", "no", "false"]:
-                        raise FirewallError(errors.INVALID_VALUE,
-                                            "'%s' for %s" % \
-                                            (new_value, property_name))
-                if property_name == "AllowZoneDrifting":
-                    if new_value.lower() not in ["yes", "true", "no", "false"]:
-                        raise FirewallError(errors.INVALID_VALUE,
-                                            "'%s' for %s" % \
-                                            (new_value, property_name))
+                elif property_name == "IPv6_rpfilter2":
+                    if new_value not in config.IPV6_RPFILTER_VALUES:
+                        raise FirewallError(
+                            errors.INVALID_VALUE,
+                            "'%s' for %s" % (new_value, property_name),
+                        )
+                else:
+                    raise dbus.exceptions.DBusException(
+                        "org.freedesktop.DBus.Error.InvalidArgs: "
+                        "Property '%s' does not exist" % property_name)
 
                 self.config.get_firewalld_conf().set(property_name, new_value)
                 self.config.get_firewalld_conf().write()
                 self.PropertiesChanged(interface_name,
                                        { property_name: new_value }, [ ])
-            elif property_name in ["MinimalMark", "AutomaticHelpers"]:
+            elif property_name in ["MinimalMark", "AutomaticHelpers", "AllowZoneDrifting"]:
                 # deprecated fields. Ignore setting them.
                 pass
             else:
@@ -781,7 +799,7 @@ class FirewallDConfig(slip.dbus.service.Object):
         log.debug1("config.PropertiesChanged('%s', '%s', '%s')",
                    interface_name, changed_properties, invalidated_properties)
 
-    @slip.dbus.polkit.require_auth(config.dbus.PK_ACTION_INFO)
+    @dbus_polkit_require_auth(config.dbus.PK_ACTION_INFO)
     @dbus_service_method(dbus.INTROSPECTABLE_IFACE, out_signature='s')
     @dbus_handle_exceptions
     def Introspect(self, sender=None): # pylint: disable=W0613
@@ -789,8 +807,16 @@ class FirewallDConfig(slip.dbus.service.Object):
 
         data = super(FirewallDConfig, self).Introspect(self.path,
                                                        self.busname.get_bus())
-        return dbus_introspection_add_properties(
-            self, data, config.dbus.DBUS_INTERFACE_CONFIG)
+        data = dbus_introspection_add_properties(
+                    self, data, config.dbus.DBUS_INTERFACE_CONFIG)
+
+        for interface in [config.dbus.DBUS_INTERFACE_CONFIG_DIRECT]:
+            data = dbus_introspection_add_deprecated(
+                        self, data, interface,
+                        dbus_service_method_deprecated().deprecated,
+                        dbus_service_signal_deprecated().deprecated)
+
+        return data
 
     # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
 
@@ -809,7 +835,7 @@ class FirewallDConfig(slip.dbus.service.Object):
     def setLockdownWhitelist(self, settings, sender=None): # pylint: disable=W0613
         log.debug1("config.policies.setLockdownWhitelist(...)")
         settings = dbus_to_python(settings)
-        self.config.get_policies().lockdown_whitelist.import_config(settings)
+        self.config.get_policies().lockdown_whitelist.import_config(settings, {})
         self.config.get_policies().lockdown_whitelist.write()
         self.LockdownWhitelistUpdated()
 
@@ -1404,6 +1430,7 @@ class FirewallDConfig(slip.dbus.service.Object):
     # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
     # DIRECT
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          out_signature=Direct.DBUS_SIGNATURE)
     @dbus_handle_exceptions
@@ -1412,6 +1439,7 @@ class FirewallDConfig(slip.dbus.service.Object):
         log.debug1("config.direct.getSettings()")
         return self.config.get_direct().export_config()
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature=Direct.DBUS_SIGNATURE)
     @dbus_handle_exceptions
@@ -1419,10 +1447,11 @@ class FirewallDConfig(slip.dbus.service.Object):
         # returns list ipv, table, list of chains
         log.debug1("config.direct.update()")
         settings = dbus_to_python(settings)
-        self.config.get_direct().import_config(settings)
+        self.config.get_direct().import_config(settings, {})
         self.config.get_direct().write()
         self.Updated()
 
+    @dbus_service_signal_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus.service.signal(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_handle_exceptions
     def Updated(self):
@@ -1430,6 +1459,7 @@ class FirewallDConfig(slip.dbus.service.Object):
 
     # chain
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='sss')
     @dbus_handle_exceptions
@@ -1449,6 +1479,7 @@ class FirewallDConfig(slip.dbus.service.Object):
         settings[0].append(idx)
         self.update(settings)
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='sss')
     @dbus_handle_exceptions
@@ -1468,6 +1499,7 @@ class FirewallDConfig(slip.dbus.service.Object):
         settings[0].remove(idx)
         self.update(settings)
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='sss', out_signature='b')
     @dbus_handle_exceptions
@@ -1480,6 +1512,7 @@ class FirewallDConfig(slip.dbus.service.Object):
         idx = tuple((ipv, table, chain))
         return idx in self.getSettings()[0]
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='ss', out_signature='as')
     @dbus_handle_exceptions
@@ -1493,6 +1526,7 @@ class FirewallDConfig(slip.dbus.service.Object):
                 ret.append(idx[2])
         return ret
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='', out_signature='a(sss)')
     @dbus_handle_exceptions
@@ -1502,6 +1536,7 @@ class FirewallDConfig(slip.dbus.service.Object):
 
     # rule
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='sssias')
     @dbus_handle_exceptions
@@ -1523,6 +1558,7 @@ class FirewallDConfig(slip.dbus.service.Object):
         settings[1].append(idx)
         self.update(tuple(settings))
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='sssias')
     @dbus_handle_exceptions
@@ -1544,6 +1580,7 @@ class FirewallDConfig(slip.dbus.service.Object):
         settings[1].remove(idx)
         self.update(tuple(settings))
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='sssias', out_signature='b')
     @dbus_handle_exceptions
@@ -1558,6 +1595,7 @@ class FirewallDConfig(slip.dbus.service.Object):
         idx = (ipv, table, chain, priority, args)
         return idx in self.getSettings()[1]
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='sss')
     @dbus_handle_exceptions
@@ -1574,6 +1612,7 @@ class FirewallDConfig(slip.dbus.service.Object):
                 settings[1].remove(rule)
         self.update(tuple(settings))
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='sss', out_signature='a(ias)')
     @dbus_handle_exceptions
@@ -1589,6 +1628,7 @@ class FirewallDConfig(slip.dbus.service.Object):
                 ret.append((idx[3], idx[4]))
         return ret
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='', out_signature='a(sssias)')
     @dbus_handle_exceptions
@@ -1598,6 +1638,7 @@ class FirewallDConfig(slip.dbus.service.Object):
 
     # passthrough
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='sas')
     @dbus_handle_exceptions
@@ -1616,6 +1657,7 @@ class FirewallDConfig(slip.dbus.service.Object):
         self.update(settings)
 
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='sas')
     @dbus_handle_exceptions
@@ -1633,6 +1675,7 @@ class FirewallDConfig(slip.dbus.service.Object):
         settings[2].remove(idx)
         self.update(settings)
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='sas', out_signature='b')
     @dbus_handle_exceptions
@@ -1644,6 +1687,7 @@ class FirewallDConfig(slip.dbus.service.Object):
         idx = (ipv, args)
         return idx in self.getSettings()[2]
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='s', out_signature='aas')
     @dbus_handle_exceptions
@@ -1656,6 +1700,7 @@ class FirewallDConfig(slip.dbus.service.Object):
                 ret.append(idx[1])
         return ret
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          out_signature='a(sas)')
     @dbus_handle_exceptions
